//////////////////////////////////////////////////////////////////////////////
//                                                                          
//  NAME: Joon S. Lee (jl)
//  FILE: PCBottomUpStmt.C
//  PROJ: COKO
//  DATE: 6/15/97
//
// SYNOPSIS
//  Method definitions for CBottomUpStmt class.
//  
//////////////////////////////////////////////////////////////////////////////

#include "PCBottomUpStmt.H"

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: CBottomUpStmt
// Parameters: (CStmt *)
//
// Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

CBottomUpStmt::CBottomUpStmt(CStmt * stmt)
{
  _stmt = stmt;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: ~CBottomUpStmt
// Parameters:	(void)
//
// Destructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

CBottomUpStmt::~CBottomUpStmt(void)
{
  if(_stmt) { delete _stmt; _stmt = NULL; }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: Exec
// Parameters:	(CState*)
//
// What to do when executed.
// Execute _stmt on Every node of the tree (current tree.)
// Bottom Up Traversal.
//         								    
//////////////////////////////////////////////////////////////////////////////

CState *CBottomUpStmt::Exec(CState *s)
{
  NonType * root = s->Store()->Copy_kTree();
  NonType * curr_child = NULL;
  int curr_which = 1;
  int success = 0;
  
  root->initCurrChild();
  curr_child = root->nextChild();

  while (curr_child)
	{
	  curr_child = curr_child->Copy_kTree();
	  s->SetStore(curr_child);

	  Exec(s); // recursion
	  success = s->ISuccess() || success;

	  NonType* result_child = s->Store()->Copy_kTree();
	  root->replaceChild(result_child, curr_which);
	  curr_child = root->nextChild();
	  curr_which++;
	}

  s->SetStore(root); // Restore old root
  

  s = _stmt->Exec(s);
  
  // set success value.
  s->SetSuccess(s->ISuccess() || success);

  return s;
}


