//////////////////////////////////////////////////////////////////
//
//  NAME: Joon S. Lee (jl), Kee-Eung Kim (kek)
//  
//  FILE: PCPattern.C
//  PROJ: COKO
//  DATE: 6/96
//  
//  SYNOPSIS
//  This class is for Match Statements
//
/////////////////////////////////////////////////////////////////

#include "PCPattern.H"
#include <KOBVar.H>
#include <KOPVar.H>
#include <KOFVar.H>
#include <KOOVar.H>

//////////////////////////////////////////////////////////////////
//
//  Function NAME: CPattern
//  Parameters: (NonType*)
//  
//  Constructor.
//
/////////////////////////////////////////////////////////////////

CPattern::CPattern(NonType* nt)
{
  rrclass = nt;
  _savedStackTop = NULL;
}

//////////////////////////////////////////////////////////////////
//
//  Function NAME: ~CPattern
//  Parameters: None.
//  
//  Destructor.
//
/////////////////////////////////////////////////////////////////

CPattern::~CPattern()
{
  if (rrclass) {delete rrclass; rrclass = NULL;}
  _savedStackTop = NULL;
}


//////////////////////////////////////////////////////////////////
//
//  Function NAME: match
//  Parameters: (CState*)
//
//  This method is for rule matching only.
//  Rule matching is different from statement matching.
//  Duplicated variable entry is allowed here.
//  
//  This is a method in CPattern class.
//  This takes a CState as an input and outputs a CState
//
//  rrclass is a field in CPattern class. It is NonType.
//  All it does is calling unify methods in NonType classes.
//
/////////////////////////////////////////////////////////////////

CState* CPattern::match(CState* s)
{
  NameUnion* tmp_name = rrclass->Myname();
  if (((tmp_name->tag == BNAME) && (tmp_name->bn == BOOL_VAR))
	  || ((tmp_name->tag == FNAME) && (tmp_name->fn == FUN_VAR))
	  || ((tmp_name->tag == PNAME) && (tmp_name->pn == PRED_VAR))
	  || ((tmp_name->tag == ONAME) && (tmp_name->on == OBJ_VAR)))
	{

	  if(_debug_) {
	  cout << "% match : CPattern" << endl;
	  cout << "----- Head is Var -----" << endl;
	  }

	  if (tmp_name->bn == BOOL_VAR)
		{ delete tmp_name; tmp_name = NULL;  return ((BVar*)rrclass)->directStore(s); 
		  }
	  if (tmp_name->fn == FUN_VAR)
		{ delete tmp_name; tmp_name = NULL; return ((FVar*)rrclass)->directStore(s); 
		  }
	  if (tmp_name->pn == PRED_VAR)
		{ delete tmp_name; tmp_name = NULL;  return ((PVar*)rrclass)->directStore(s); 
		  }
	  if (tmp_name->on == OBJ_VAR)
		{ delete tmp_name; tmp_name = NULL; return ((OVar*)rrclass)->directStore(s); 
		  }
	}
  else
	{  delete tmp_name; tmp_name = NULL; return rrclass->unify(s); }
}

//////////////////////////////////////////////////////////////////
//
//  Function NAME: matchSTM
//  Parameters: (CState*)
//  
//  This match method is used for matching statement.
//  No duplicated variable entry is allowed.
//  Added on March 19 by Joon S. Lee
//
//  This is a method in CPattern class.
//  This takes a CState as an input and outputs a CState
//
//  rrclass is a field in CPattern class. It is NonType.
//  All it does is calling unify methods in NonType classes.
//
/////////////////////////////////////////////////////////////////
CState* CPattern::matchSTM(CState* s)
{
  NameUnion* tmp_name = rrclass->Myname();
  if (((tmp_name->tag == BNAME) && (tmp_name->bn == BOOL_VAR))
	  || ((tmp_name->tag == FNAME) && (tmp_name->fn == FUN_VAR))
	  || ((tmp_name->tag == PNAME) && (tmp_name->pn == PRED_VAR))
	  || ((tmp_name->tag == ONAME) && (tmp_name->on == OBJ_VAR)))
	{

	  if(_debug_) {
	  cout << "% match : CPattern" << endl;
	  cout << "----- Head is Var -----" << endl;
	  }

	  if (tmp_name->bn == BOOL_VAR)
		{ delete tmp_name; tmp_name = NULL;  return ((BVar*)rrclass)->directStore(s); 
		  }
	  if (tmp_name->fn == FUN_VAR)
		{ delete tmp_name; tmp_name = NULL; return ((FVar*)rrclass)->directStore(s); 
		  }
	  if (tmp_name->pn == PRED_VAR)
		{ delete tmp_name; tmp_name = NULL;  return ((PVar*)rrclass)->directStore(s); 
		  }
	  if (tmp_name->on == OBJ_VAR)
		{ delete tmp_name; tmp_name = NULL; return ((OVar*)rrclass)->directStore(s); 
		  }
	}
  else
	{  delete tmp_name; tmp_name = NULL; return rrclass->unifySTM(s); }
}


//////////////////////////////////////////////////////////////////
//
//  Function NAME: build
//  Parameters: (CState*)
//  
//  This calls transform methods.
//
/////////////////////////////////////////////////////////////////

CState* CPattern::build(CState* s)
{
  NonType* a;
  NonType* b;

  b = (s->Store())->Copy_kTree();

  if(_debug_)
	cout << "%CPattern : build" << endl;

  a = rrclass->transform(s);

  // transform succeeded, return the new transformed NonType

  if (s->ISuccess())
  {
	delete b; b = NULL;
	s->SetStore(a);
	s->SetSuccess(T);
	return s;
  }
  // transform failed, return the old NonType
  else
  {
	delete a; a = NULL;
	s->SetStore(b);
	s->SetSuccess(F);
	return s;
  }
}

//////////////////////////////////////////////////////////////////
//
//  Function NAME: DepListCheck
//  Parameters: (char*, CState*)
//  
//
/////////////////////////////////////////////////////////////////
CState* CPattern::DepListCheck(char* vn, CState* s)
{
  if (_debug_)
	cout << "DepListCheck@CPattern starts" << endl;

  if (vn)
	{
	  if (!s->IFindDeep(vn))
		{
		  cout << "Error @ CPattern : DepListCheck" << endl;
		  cout<< "Variable Not Initialized (redefine uninitialized variable)" << endl;
		  s->SetSuccess(F);
		  exit(1);
		}
	}

  if (_debug_)
	cout << "******************" << endl << "rrclass->setDEPnInvDEP @CPattern starts" << endl;  
  
  rrclass->setDEPnInvDEP(vn,s);
  
  if (_debug_)
	cout << "******************" << endl << "rrclass->setDEPnInvDEP @CPattern ends" << endl;
  
  s->SetSuccess(T);
  
  if (_debug_)
	cout << "DepListCheck@CPattern ends" << endl;

  return s;
}

//////////////////////////////////////////////////////////////////
//
//  Function NAME: undoDEP
//  Parameters: (char*, CState*)
//  
//
/////////////////////////////////////////////////////////////////
CState* CPattern::undoDEP(char* vn, CState* s)
{
  rrclass->undoSetInvDEP(vn,s);
  return s;
}


//////////////////////////////////////////////////////////////////
//
//  Function NAME: buildFromHere
//  Parameters: (CState*, CIEnvStackDataType*)
//  
//  This calls transformFromHere methods.
//
/////////////////////////////////////////////////////////////////

CState* CPattern::buildFromHere(CState* s, CIEnvStackDataType* startingPt)
{
  NonType* a;
  NonType* b;

  b = (s->Store())->Copy_kTree();
  a = rrclass->transformDeep(s, startingPt);

  if (s->ISuccess())
  {
	delete b; b = NULL;
	s->SetStore(a);
	s->SetSuccess(T);
	return s;
  }
  else
  {
	delete a; a = NULL;
	s->SetStore(b);
	s->SetSuccess(F);
	return s;
  }
}


//////////////////////////////////////////////////////////////////
//
//  Function NAME: buildCurrent
//  Parameters: (CState*)
//  
//  This calls buildFromHere methods.
//
/////////////////////////////////////////////////////////////////

CState* CPattern::buildCurrent(CState* s) 
{ 
  if (_savedStackTop)
	return buildFromHere(s, _savedStackTop); 
  else
	return buildFromHere(s, s->IEnv()->getData());
}


//////////////////////////////////////////////////////////////////
//
//  Function NAME: SetStackPointer
//  Parameters: (CIEnvStackDataType*)
//  
//
/////////////////////////////////////////////////////////////////

void CPattern::SetStackPointer(CIEnvStackDataType* stackPt)
{ 
  _savedStackTop = stackPt; 
}
