//////////////////////////////////////////////////////////////////////////////
//                                                                          
//  NAME: Joon-Suk Lee (jl)
//  FILE: PCPreCond.C
//  PROJ: COKO
//  DATE: 10/9/96                                                              
//
// SYNOPSIS
// Class definitions for CPreCond.
//  
//////////////////////////////////////////////////////////////////////////////

#ifndef _PCMULMATCHSTMT_C_
#define _PCMULMATCHSTMT_C_

#include "PCPreCond.H"
#include "CGlobals.H"


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: CPreCond
// Parameters:	(char* a ... )
//
// Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

CPreCond::CPreCond(char* a ... )
{
  va_list ap;
  va_start (ap,a);
  _RuleBlockName = new char[strlen(a)+1];
  _condList = new CondList();
  strcpy(_RuleBlockName, a);

   _id = va_arg(ap,int);

  for(;;)
	{
	  CCond* p = va_arg(ap, CCond*);
	  if (p == NULL) 
		break;
	  else 
		_condList->insert(p);
	}
  va_end(ap);
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: ~CPreCond
// Parameters:	(void)
//
// Destructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

CPreCond::~CPreCond(void)
{
  if (_condList) { delete _condList; _condList = NULL; }

  if (_RuleBlockName) {delete _RuleBlockName; _RuleBlockName = NULL;}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: Exec
// Parameters:	(CState*)
//
// What to do when executed.
//         								    
//////////////////////////////////////////////////////////////////////////////

CState* CPreCond::Exec(CState * s, NonType* nt)
{
#ifdef _PROLOG_

  if (_debug_)
 	cout << "-- PreCond Begins" << endl;
  
  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////
  
  // btr_schema file
  {
	fstream pl_file, ql_file;

 	char* pl_file_name = "btr_schema.pl";
	char* ql_file_name = "btr_schema.ql";

 	struct stat sbuf1;
 	struct stat sbuf2;
 	
	stat(pl_file_name,&sbuf1);
 	stat(ql_file_name,&sbuf2);

 	pl_file.open(pl_file_name, ios::in);
 	ql_file.open(ql_file_name, ios::in);

 	if (!pl_file)
 	{
	  // CASE 0: When btr_schema.pl does not exist.
 	  // This is an error.
 	  cerr << "Error - btr_schema.pl file does not exist" << endl;
 	  exit(-1);
 	}
 	else if (!ql_file)
 	  {
		// CASE 1: When btr_schema.ql files does not exist.
		// Compile btr_schema.pl into btr_schema.ql.

 		pl_file.close();

		int strlength = strlen("/pro/oodb/cokoc/prolog/sicstus3/inst_dir/bin/sicstus ") + strlen("-l ") 
		  + strlen("fcompile_btr_schema.pl") + 1;
		char* buffer = new char[strlength];
		strcpy(buffer, "/pro/oodb/cokoc/prolog/sicstus3/inst_dir/bin/sicstus -l fcompile_btr_schema.pl");
		system(buffer);
		delete [] buffer;
 	  }
 	else if (sbuf1.st_mtime > sbuf2.st_mtime)
 	  {
 		// CASE 2: When btr_schema.pl and btr_schema.ql do exist.
 		// In this case, compare time stamps for those files.
 		//
 		// If ql file is created before the time pl file is last modified,
 		// recompile pl file into new ql file.

		pl_file.close();
 		ql_file.close();

		int strlength = strlen("/pro/oodb/cokoc/prolog/sicstus3/inst_dir/bin/sicstus ") + strlen("-l ") 
		  + strlen("fcompile_btr_schema.pl") + 1;
		char* buffer = new char[strlength];
		strcpy(buffer, "/pro/oodb/cokoc/prolog/sicstus3/inst_dir/bin/sicstus -l fcompile_btr_schema.pl");
		system(buffer);
		delete [] buffer;

 		if (_debug_)
 		  {
 			cout << "pl_file :- " << asctime(localtime(&sbuf1.st_mtime));
 			cout << "ql_file :- " << asctime(localtime(&sbuf2.st_mtime));
 		  }

 	  }
  }

  // _fin file
  {
	fstream pl_file, ql_file;

	char* pl_file_name = new char[strlen(_RuleBlockName) + strlen("_fin.pl") + 1];
	sprintf(pl_file_name, "%s_fin.pl", _RuleBlockName);
	
	char* ql_file_name = new char[strlen(_RuleBlockName) + strlen("_fin.ql") + 1];
	sprintf(ql_file_name, "%s_fin.ql", _RuleBlockName);

 	struct stat sbuf1;
 	struct stat sbuf2;
 	
	stat(pl_file_name,&sbuf1);
 	stat(ql_file_name,&sbuf2);

 	pl_file.open(pl_file_name, ios::in);
 	ql_file.open(ql_file_name, ios::in);

 	if (!pl_file)
 	{
	  // CASE 0: When pl file does not exist.
 	  // This is an error.
 	  cerr << "Error - " << _RuleBlockName << ".pl file does not exist" << endl;
 	  exit(-1);
 	}
 	else if (!ql_file)
 	  {
		// CASE 1: When ql files does not exist.
		// Compile pl file into ql file.

 		pl_file.close();

		int strlength = strlen("/pro/oodb/cokoc/prolog/sicstus3/inst_dir/bin/sicstus ") + strlen("-l ") 
		  + strlen("fcompile__fin.pl") + strlen(_RuleBlockName) + 1;
		char* buffer = new char[strlength];
		sprintf(buffer, "/pro/oodb/cokoc/prolog/sicstus3/inst_dir/bin/sicstus -l fcompile_%s_fin.pl", _RuleBlockName);
		system(buffer);
		delete [] buffer;
 	  }
 	else if (sbuf1.st_mtime > sbuf2.st_mtime)
 	  {
 		// CASE 2: When pl and ql files do exist.
 		// In this case, compare time stamps for those files.
 		//
 		// If ql file is created before the time pl file is last modified,
 		// recompile pl file into new ql file.

		pl_file.close();
 		ql_file.close();

		int strlength = strlen("/pro/oodb/cokoc/prolog/sicstus3/inst_dir/bin/sicstus ") + strlen("-l ") 
		  + strlen("fcompile__fin.pl") + strlen(_RuleBlockName) + 1;
		char* buffer = new char[strlength];
		sprintf(buffer, "/pro/oodb/cokoc/prolog/sicstus3/inst_dir/bin/sicstus -l fcompile_%s_fin.pl", _RuleBlockName);
		system(buffer);
		delete [] buffer;

 	  }
	if (pl_file_name) {delete pl_file_name; pl_file_name = NULL;}
	if (ql_file_name) {delete ql_file_name; ql_file_name = NULL; }
  }
  
  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////
  
  
  NonType* _nt = nt;

  // calling exec methods for all CCond objects.

  _condList->initIterator();

  CCond* c = _condList->getValue();

  while(c)
	{
	  s = c->Exec(s);
	  c = _condList->getValue();
	}

  // invoke prolog query 
  
  SP_pred_ref pred;
  SP_term_ref input_term, var_term;

  char* result_text = new char [3000]; // the result from the prolog gets saved here.

  if (!is_prolog_initialized)
	{
	  SP_initialize(0, NULL, NULL);
	  is_prolog_initialized = 1;
	}

  char* ql_file_name = new char[strlen(_RuleBlockName) + strlen("_fin.ql") + 1];
  sprintf(ql_file_name,"%s_fin.ql",_RuleBlockName);

  int load_result = SP_load(ql_file_name); // SP_load
//   load_result = SP_load(ql_file_name); // SP_load
//   load_result = SP_load(ql_file_name); // SP_load
//   load_result = SP_load(ql_file_name); // SP_load
//   load_result = SP_load(ql_file_name); // SP_load
//   load_result = SP_load(ql_file_name); // SP_load
//   load_result = SP_load(ql_file_name); // SP_load
//   load_result = SP_load(ql_file_name); // SP_load

  // if (load_result)
  //cout << endl << endl << ql_file_name << " loaded ok." << endl << endl;
  //  else
  //	cout << endl << endl << ql_file_name << " did not loaded ok." << endl << endl;

  if(ql_file_name) { delete [] ql_file_name; ql_file_name = NULL;}

  char* id_char = new char[5];  // convert _id(int) to string(char*)

  sprintf(id_char,"%d",_id);

  char* query_name = new char[strlen("query") + strlen(id_char) + 1];
  sprintf(query_name,"query%s",id_char);
  
  pred = SP_predicate(query_name,2,""); // SP_predicate
  
  if (id_char) { delete [] id_char; id_char = NULL;}
  if (query_name) { delete [] query_name; query_name = NULL; }

  input_term = _nt->sp_prolog();  // prepare for the query input.
  // cout << "In PCPreCond: Calling prolog () on query leaves: " << _nt->prolog() << endl;
  var_term = SP_new_term_ref();
  SP_put_variable(var_term);
  
  
  if (SP_query(pred, input_term, var_term))
	{
	  s->SetSuccess(T); 
	  SP_get_list_chars(var_term,&result_text);
	}
  else
	s->SetSuccess(F);

  // parsing the results.
  //    cout << "result_text " << result_text <<  endl;
  
  if (s->ISuccess())
 	{
	  _condList->initIterator();
	  c = _condList->getValue();
	  
 	  PrologParser* _parList[MAXARG];
 	  int parser_count = 0;

	  PrologParser *p = new PrologParser();
	  
	  
	  char* tmp_result = strtok(result_text," ");
	  //	  char* tmp_result =  new char[strlen(tmp)+1];

	  do
 		{
		  //		  cout << " tmp result " << tmp_result << endl;
		  if (tmp_result)
			{
			  p->setInput(tmp_result);
			  NonType* k = p->parse();
			  
			  while (c && c->allDone())
				c = _condList->getValue();
			  
			  if (c)
				s == c->build(s, k);
			  
			  while (c && c->allDone())
				c = _condList->getValue();
			  
			  if (k) {delete k; k = NULL;}
			  
			  tmp_result = strtok(NULL," "); // Joon:  Can you explain this to me?

			  //if (tmp_result) { delete tmp_result; tmp_result = NULL; }
			  //tmp_result = new char[strlen(tmp)+1];
			  //		  sprintf(tmp_result, "%s", tmp);
			}
		  else
			{
			  // do none
			  // call to prolog was successful but did not return any 
			  // variable to kola tree bindings.
			}
		  
  		} while (tmp_result);
	  
	  if (p) {delete p; p = NULL;}
	  
 	}
  
  // mfc:  This seems to be deleting environment entries:  if (result_text) {delete [] result_text; result_text = NULL;}
  
  if (_debug_)
  	cout << "-- PreCond Ends" << endl;
  
#endif // _PROLOG_
  return s;  
}

#endif

