/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//                                                                          
//  NAME: Mitch Cherniack (mfc)					 	    
//  FILE: ENEnvironment.H 
//  PROJ: The "Karbonator"
//  DATE: 5/95		                                                    
//
// SYNOPSIS
//  Environment Object.  Used to Pass Information about lambda bound
//   variables down tree during AQUA -> KOLA translation.
//  
//////////////////////////////////////////////////////////////////////////////

#ifndef KOLA_ENENVIRONMENT_HEADER
#define KOLA_ENENVIRONMENT_HEADER 

#include "../KGlobals.H"
#include "ENEnvEntry.H"
#include "KOFunction.H"
#include "KOF1.H"
#include "KOF2.H"
#include "KOFID.H"
#include "KOFCompose.H"
#include "KOFPlus.H"
#include "KOFIterate.H"
#include "KOPConst.H"
#include "KOBPrimitive.H"
#include "KOFChoose.H"
#include "TYTAny.H"
#include "TYTObj.H"
#include "TYTSet.H"
#include "TYTPair.H"
#include "TYTTup.H"

#include <string.h>
#include <iostream.h>

class Environment 
{
    
 public:
  Environment ();
  // Constructor
    
  Environment (Environment& other);
  // Copy Constructor
    
  ~Environment ();
  // Destructor
    
  char* show ();
  //  For debugging purposes.  Show current entries in table

  int isEmpty () {return fst == NULL;}
  //  Is the environment currently empty?

  EnvEntry *Top () {return fst;}
  // Selector.
    
  EnvEntry *Last () {return lst;}
  // Selector.
    
  Function* GetFunc (char *var);
  // Return function that gets var by looking up in environment

  TObj* GetType (char *var);
  // Return type associated with var by looking up in environment

  Environment *add_entry (char *var, TObj *t);
  // Adds (var, _2) to top of environment, and adds o _1 to every other entry

  Environment *add_pair_entry (char *var1, char *var2, TObj *t1, TObj *t2);
  // Adds (var1, _1 o _2), (var2, _2 o _2) to top of environment, and adds o _1 to every other entry

  Environment *add_manual_entry (char *var, TObj *ty);
  // Add manual entry to table.  Used for environment as schema manager.

  Environment *add_manual_entry (char *var, Function *f, TObj *ty);
  // Add manual entry to table.  Used for environment as label manager

  Environment *insert (EnvEntry *ent);
  // Insert ent into environment

  EnvEntry *reversed ();
  // Return chain of copies of EnvEntries in chain buit in reversed order

  Function *buildtup_unnest (char *fn);
  // Build a function that constucts tuple denoted in environment, but
  // replacing function to access fn with f
    
  Function *buildtup_nest (char *fn);
  // Build a function that constucts tuple denoted in environment, but
  // replacing function to access fn with f
    
  Function *buildtup_without (char *fn);
  // Build a function that constucts tuple denoted in environment, but
  // without function accessing label fn
    
  TObj *buildtuptype_unnest (char *fn);
  // Build type denoting type of tuple represented by environment, but unnested on label fn
    
  TObj *buildtuptype_nest (char *fn);
  // Build type denoting type of tuple represented by environment, but unnested on label fn

  int sameas (Environment *other);
  // return true if this environment has same names/types as other
    
 protected:
  EnvEntry *fst;
  // First entry in environment list

  EnvEntry *lst;
  // Last entry in environment list
    
};


#endif


