/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Mitch Cherniack (mfc@cs.brown.edu)
//       Joon S. Lee (jl@cs.brown.edu)
// FILE: KOFFold.C
// PROJ: COKO Compiler, KOLA Parser
// DATE: 5/97
//
//////////////////////////////////////////////////////////////////////////////

#include "KOFFold.H"


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  FFold
// Parameters:	(koObject *o, Function *f, Function *op)					   
//
//  Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

FFold::FFold (koObject *o, Function *f, Function *op, TFunc * ftype)
                                : Function (ftype, FUN_FOLD,"FOLD")
{
  one = o;
  two = f;
  three = op;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  FFold
// Parameters:	(FFold& other)							  
//
//  Copy Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

FFold::FFold (const FFold& other) : Function ((Function &) other)
{
  one = (other . One ()) -> Copy ();
  two = (other . Two ()) -> Copy ();
  three = (other . Three ()) -> Copy ();
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~FFold
// Parameters:	()							  
//
//  Destructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

FFold::~FFold ()
{
  delete one;
  delete two;
  delete three;
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Copy ()
// Parameters:	()							  
//
//  Virtual Copy Constructor 
//         								    
//////////////////////////////////////////////////////////////////////////////

Function *FFold::Copy () const
{
  return new FFold (*this);
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  deepeq
// Parameters:  ()
//
// Value-based equality.
// this is computed recursively.
// Return true if two trees have identical structures, otherwise return false.
// No unification occurs here.
//
//////////////////////////////////////////////////////////////////////////////

int FFold::deepeq (NonType * other) 
{
  NameUnion* name_ = other->Myname();
  if ((name_->tag == FNAME ) && (name_->fn == FUN_FOLD))
  {
    delete name_;
    return ((one->deepeq(((FFold*)other)->One())) && 
            (two->deepeq(((FFold*)other)->Two())) && 
            (three->deepeq(((FFold*)other)->Three())));
  }
  else
    {
      delete name_;
      return F;
    }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Myname
// Parameters:	()							  
//
//  Returns a union type, NameUnion. 
//  Since NonType has 5 different sub-classes, Function, Predicate, Object, 
//  Bool and Index, and those 4 different sub-classes use different 
//  enum type for their own sub-classes' names, Myname methods needs 
//  a union type of 5 different enum types.
//  Since Myname is defined virtually in KONonType, one can just call this 
//  methods with NonType* to identify the specific instance of Kola object.
//         								    
//////////////////////////////////////////////////////////////////////////////

NameUnion* FFold::Myname()
{
	NameUnion* a = new NameUnion(FNAME, FUN_FOLD);
	return a;
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  nextChild
// Parameters:  ()
//
// Returns "currChild++"th child.
// When currChild is equal to the number of children
// (when current child is the last child), it returns NULL.
//
//////////////////////////////////////////////////////////////////////////////

NonType* FFold::nextChild()
{
  if (currChild == 0)
	{
	  currChild++;
	  return one;
	}
  else if (currChild == 1)
	{
	  currChild++;
	  return two;
	}
  else if (currChild == 2)
	{
	  currChild++;
	  return three;
	}
  else return NULL;
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  reduce
// Parameters:	(OResult* a)
//  return
//    new OResBB ("fold (" + one -> reduce () + ", " +
//                           two -> representation () + ", " +
//  			 three -> representation () + ")",
//  		a)
//
//////////////////////////////////////////////////////////////////////////////

// OResult* FFold::reduce (OResult* a)
// {
//   char *or = one -> representation ();
//   char *tr = two -> representation ();
//   char *thr = three -> representation ();

//   char *bbox = new char [strlen (or) + strlen (tr) + strlen (thr) + strlen ("FOLD (, , )") + 1];
//   sprintf (bbox, "FOLD (%s, %s, %s)", or, tr, thr);

//   OResult* res = new OResBB (bbox, a -> Copy ());
//   delete [] or;
//   delete [] tr;
//   delete [] thr;
//   delete [] bbox;
//   delete a;
//   return res;
// }

// //////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  representation
// Parameters:	()							  
//
//  "FFO: FFold (" + one -> representation () + two -> rep () + thr + rep () + ")"
//         								    
//////////////////////////////////////////////////////////////////////////////

char* FFold::representation ()
{
  char *or = one -> representation ();
  char *tr = two -> representation ();
  char *thr = three -> representation ();
  char *res = new char [strlen (or) + strlen (tr) + strlen (thr) + strlen ("FOLD (, , )") + 1];
  sprintf (res, "FOLD (%s, %s, %s)", or, tr, thr);
  delete [] or;    
  delete [] tr;    
  delete [] thr;    
  return res;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transform
// Parameters:	(CState*)							  
//
//  Return coresponding Nontype object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//
//////////////////////////////////////////////////////////////////////////////

NonType* FFold::transform(CState* s)
{
  koObject* first = (koObject*) One() -> transform(s);
  Function* second = (Function*) Two() -> transform(s);
  Function* third = (Function*) Three() -> transform(s);
  return new FFold(first,second,third,Type());
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  transformDeep
// Parameters:  (CState*, CIEnvStackDataType*)
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly.
//  The difference between this method and transform method is a search
//  method used in VAR classes. (Take a look at transform and transformDeep
//  methods in BVar, FVar, PVar, or OVar.)
//  Transform method searches the top frame only,
//  while TransformDeep methods searches entire stack starting from
//  a given point.
//
//////////////////////////////////////////////////////////////////////////////

NonType* FFold::transformDeep(CState* s, CIEnvStackDataType* startingPt)
{
  koObject* first = (koObject*) One() -> transformDeep(s, startingPt);
  Function* second = (Function*) Two() -> transformDeep(s, startingPt);
  Function* third = (Function*) Three() -> transformDeep(s, startingPt);
  return new FFold(first,second,third,Type());
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: compares
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for rule matching only.
// Rule matching is different from statement matching.
// Duplicated variable entry is allowed here.
//
// For example, foo: g o g --> g o g is a legal rule.
//                  -------   -------
//                    |         |_\
//                    |_\         / building part
//                      / matching part
//
// Assume that we apply this rule to a kola tree of the form "ID o ID."
// When unification occurs in a matching part, first we store first g and ID pair
// into the env stack. Then the next time we face another g (second one),
// we search the stack for the matching pair with variable name g and simply check
// if the matching pair in the stack has same structure as the matching pair we have.
// In this case, second pair is (g, ID) as well and the rule gets fired successfully.
//
// Now assume that we apply the same rule to a kola tree of the form "ID o (ID o ID)."
// In this case, the rule fails since the second pair is (g, ID o ID),
// while the first pair is (g, ID).
//
// Note that even in a case of failure, the rule itself is legal.
//
// "Duplication of a variable is legal in a rule section."
//
//////////////////////////////////////////////////////////////////////////////

int FFold::compares(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();
  
  if ((tmp_name->tag == FNAME) && (tmp_name->fn == FUN_FOLD))
	{
	  // do it for first child.
	  if (one->compares(((FFold*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one))
			// child (one) is a variable.
			{
			  if ((a->IFind(((OVar*)one)->Name())) == NULL)
				{
				  // child is a new variable.
				  a->IInsert(((OVar*)one)->Name(), new CIdent((FFold*)nt,((FFold*)nt)->One(),1));
				  // Push entry into the stack.
				}
			  else 
				{
				  // Child is a variable but not new one.
				  // In this case, I need to make sure that what's
				  // already in the stack and what I have now is the same.
				  if (!(((a->IFind(((OVar*)one)->Name()))->Child_Look_Up())->deepeq(((FFold*)nt)->One())))
					return_val++;
				  // match failed.
				  // else do none.
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++;
	  // compare failed
	  
	  // do it for second child.
	  if (two->compares(((FFold*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((FVar*)two)->Name())) == NULL)
				{
				  a->IInsert(((FVar*)two)->Name(), new CIdent((FFold*)nt, ((FFold*)nt)->Two(), 2));
				}
			  else 
				{
				  if (!(((a->IFind(((FVar*)two)->Name()))->Child_Look_Up())->deepeq(((FFold*)nt)->Two())))
					return_val++;
				}
			}
		}
	  else return_val++;

	  // do it for third child.
	  if (three->compares(((FFold*)nt)->Three(), a))
		{
		  if (isChildVar(three))
			{
			  if ((a->IFind(((FVar*)three)->Name())) == NULL)
				{
				  a->IInsert(((FVar*)three)->Name(), new CIdent((FFold*)nt, ((FFold*)nt)->Three(), 3));
				}
			  else 
				{
				  if (!(((a->IFind(((FVar*)three)->Name()))->Child_Look_Up())->deepeq(((FFold*)nt)->Three())))
					return_val++;
				}
			}
		}
	  else return_val++;

	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0.
  // In this case, return false(0).
  // Otherwise, return true(1).
  
  delete tmp_name;
 
  if (return_val)
        return F;
  else
        return T;

} 

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: comparesSTM
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for statement matching only.
// Statement matching is different from rule matching.
// No duplicated variable entry is allowed here.
//
// For example,
//             match f ! _O,
//                   f =  g o g  is a illegal statement (equation).
//                       -------
//                         |
//                         |_\
//                           / matching part
//
// Assume that we have a query tree of the form "ID o ID ! 3."
// Variable f gets bounded to "ID o ID" and we have to match "ID o ID" to "g o g."
// Matching "ID o ID" to "g o g" is perfectly legal in a rule section.
// However, it is illegal in statement section of the program.
// Remember that every variable on the right hand side of an equation must be a fresh one.
// However, after we match first g to ID, the second g is no longer fresh variable.
//
// "No duplication of a variable is legal in a statement section."
//
//////////////////////////////////////////////////////////////////////////////

int FFold::comparesSTM(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();
  
  if ((tmp_name->tag == FNAME) && (tmp_name->fn == FUN_FOLD))
	{
	  // do it for first child.
	  if (one->comparesSTM(((FFold*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one))
			// child (one) is a variable.
			{
			  if ((a->IFind(((OVar*)one)->Name())) == NULL)
				{
				  // child is a new variable.
				  a->IInsert(((OVar*)one)->Name(), new CIdent((FFold*)nt,((FFold*)nt)->One(),1));
				  // Push entry into the stack.
				}
			  else 
				{
				  return_val++;
				  cout << "Error : Variable " << ((OVar*)one)->Name() << " is used twice. " << endl;
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++;
	  // compare failed
	  
	  // do it for second child.
	  if (two->comparesSTM(((FFold*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((FVar*)two)->Name())) == NULL)
				{
				  a->IInsert(((FVar*)two)->Name(), new CIdent((FFold*)nt, ((FFold*)nt)->Two(), 2));
				}
			  else 
				{
				  return_val++;
				  cout << "Error : Variable " << ((FVar*)two)->Name() << " is used twice. " << endl;
				}
			}
		}
	  else return_val++;

	  // do it for third child.
	  if (three->comparesSTM(((FFold*)nt)->Three(), a))
		{
		  if (isChildVar(three))
			{
			  if ((a->IFind(((FVar*)three)->Name())) == NULL)
				{
				  a->IInsert(((FVar*)three)->Name(), new CIdent((FFold*)nt, ((FFold*)nt)->Three(), 3));
				}
			  else 
				{
					return_val++;
					cout << "Error : Variable " << ((FVar*)three)->Name() << " is used twice. " << endl;
				}
			}
		}
	  else return_val++;

	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0.
  // In this case, return false(0).
  // Otherwise, return true(1).
  
  delete tmp_name;
 
  if (return_val)
	return F;
  else
	return T;

} 

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  setDEPnInvDEP
// Parameters:  (char*, CState*)
//
// set dependency list (ancestor list) and inverse dependency list (descendant list)
//
//////////////////////////////////////////////////////////////////////////////

void FFold::setDEPnInvDEP(char* vn, CState* s)
{
  one->setDEPnInvDEP(vn,s);
  two->setDEPnInvDEP(vn,s);
  three->setDEPnInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  undoSetInvDEP
// Parameters:  (char*, CState*)
//
// undo changes made on descendant list (inverse dependency list).
// undo for dependency list (ancestor list) is not necessary, since
// any changes on dependency list is undone by poping up
// top frame from the environment stack.
//
//////////////////////////////////////////////////////////////////////////////

void FFold::undoSetInvDEP(char* vn, CState* s)
{
  one->undoSetInvDEP(vn,s);
  two->undoSetInvDEP(vn,s);
  three->undoSetInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  replaceChild
// Parameters:	(NonType*, int)
//
// Replace a child of given tree.  
// which_child specifies which child to replace.
// For example, when which_child is one, we replace the first child of 
// given tree.
// 
//////////////////////////////////////////////////////////////////////////////

void FFold::replaceChild(NonType* child, int which_child)
{
  if ( which_child == 1 )
	SetOne(child);
  else if ( which_child == 2 )
	SetTwo(child);
  else if ( which_child == 3 )
	SetThree(child);
}

#ifdef _OSTORE_
//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: invoke
//  Parameters:(koObject*)
//
//  Implements the function fold former.
//
//////////////////////////////////////////////////////////////////////////////

koObject *FFold::invoke(koObject *x)
{
  void *handle = dlopen(ffoldOpt, RTLD_LAZY);
  void *symbol = dlsym(handle, "callOpt");

  NonType *result = ((callOptPtr)symbol)(new OInvoke(this,x));
  dlclose(handle);
  return (koObject*)result;
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: exec
//  Parameters:(OSObject*)
//
//  Implements the function fold former.
//
//////////////////////////////////////////////////////////////////////////////

OSObject* FFold::exec(OSObject* x)
{
  ((Iterator*)x)->open();

  if (((Iterator*)x)->finished()) {  // empty collection
    ((Iterator*)x)->close();
    return One()->result();
  }

  else {

    OSObject *d = One()->result();
    while (!((Iterator*)x)->finished()) {

      OSObject *e = ((Iterator*)x)->next();
      OSPair *pair = new OSPair(Two()->invoke(d->kola())->result(),
                                Two()->invoke(e->kola())->result());

      d = Three()->invoke(pair->kola())->result();
    }

    ((Iterator*)x)->close();
    return d;
  }
}

#endif // _OSTORE_
  
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  prolog
// Parameters:	()
//
// return string representation in prolog term
//
//////////////////////////////////////////////////////////////////////////////

char* FFold::prolog ()
{
  char *or = one -> prolog ();
  char *tr = two -> prolog ();
  char *thr = three -> prolog ();
  char *res = new char [strlen (or) + strlen (tr) + strlen (thr) + strlen ("ffold(, , )") + 1];
  sprintf (res, "ffold(%s, %s, %s)", or, tr, thr);
  delete [] or;    
  delete [] tr;    
  delete [] thr;    
  return res;
}

#ifdef _PROLOG_
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  sp_prolog
// Parameters:	()
//
// Since C/C++ functions only have indirect access to Prolog terms 
// via SP_term_refs, we need to convert kola tree to SP_term_refs.
// Converted kola tree will get passed to prolog runtime system.
//
// See the Sicstus prolog version 3#6 manual for more detailed info on
// the usage of SP_term_refs and other prolog supplied methods.
//
//////////////////////////////////////////////////////////////////////////////

SP_term_ref FFold::sp_prolog ()
{
  SP_term_ref or = one -> sp_prolog ();
  SP_term_ref tr = two -> sp_prolog ();
  SP_term_ref thr = three -> sp_prolog ();
  SP_term_ref res = SP_new_term_ref();
  SP_cons_functor(res, SP_atom_from_string("ffold"),3,or,tr,thr);
  return res;
}
#endif // _PROLOG_
