/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Mitch Cherniack (mfc@cs.brown.edu)
//       Joon S. Lee (jl@cs.brown.edu)
// FILE: KOFIter.C
// PROJ: COKO Compiler, KOLA Parser
// DATE: 5/97
//
//////////////////////////////////////////////////////////////////////////////

#include "KOFIter.H"

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  FIter
// Parameters:	(Predicate* eq, Function *f)							  
//
//  Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

FIter::FIter (Predicate *p, Function *f, TFunc * ftype)
                            : Function (ftype, FUN_ITER,"ITER")
{
  one = p;
  two = f;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  FIter
// Parameters:	(FIter& other)							  
//
//  Copy Constructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

FIter::FIter (const FIter& other) : Function ((Function &) other)
{
  one = (other . One ()) -> Copy ();
  two = (other . Two ()) -> Copy ();
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~FIter
// Parameters:	()							  
//
//  Destructor.  
//         								    
//////////////////////////////////////////////////////////////////////////////

FIter::~FIter ()
{
  delete one;
  delete two;
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Copy ()
// Parameters:	()							  
//
//  Virtual Copy Constructor 
//         								    
//////////////////////////////////////////////////////////////////////////////

Function *FIter::Copy () const
{
  return new FIter (*this);
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  deepeq
// Parameters:  ()
//
// Value-based equality.
// this is computed recursively.
// Return true if two trees have identical structures, otherwise return false.
// No unification occurs here.
//
//////////////////////////////////////////////////////////////////////////////

int FIter::deepeq (NonType * other) 
{
  NameUnion* name_ = other->Myname();
  if ((name_->tag == FNAME ) && (name_->fn == FUN_ITER))
  {
    delete name_;
    return ((one->deepeq(((FIter*)other)->One())) 
         && (two->deepeq(((FIter*)other)->Two())));
  }
  else
    {
      delete name_;
      return F;
    }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Myname
// Parameters:	()							  
//
//  Returns a union type, NameUnion. 
//  Since NonType has 5 different sub-classes, Function, Predicate, Object, 
//  Bool and Index, and those 4 different sub-classes use different 
//  enum type for their own sub-classes' names, Myname methods needs 
//  a union type of 5 different enum types.
//  Since Myname is defined virtually in KONonType, one can just call this 
//  methods with NonType* to identify the specific instance of Kola object.
//         								    
//////////////////////////////////////////////////////////////////////////////
 
NameUnion* FIter::Myname()
{
	NameUnion* a = new NameUnion(FNAME, FUN_ITER);
	return a;
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  nextChild
// Parameters:  ()
//
// Returns "currChild++"th child.
// When currChild is equal to the number of children
// (when current child is the last child), it returns NULL.
//
//////////////////////////////////////////////////////////////////////////////

NonType* FIter::nextChild()
{
  if (currChild == 0)
	{
	  currChild++;
	  return one;
	}
  else if (currChild == 1)
	{
	  currChild++;
	  return two;
	}
  else return NULL;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  reduce
// Parameters:	()							  
//
/////////////////////////////////////////////////////////////////////////////

// OResult* FIter::reduce (OResult* a)
// {
//     if (a -> isPair ()) {

// 	OResPAIR *a2 = (OResPAIR *) a;
//         OResVAR *v = new OResVAR ();
//         OResPAIR *pr = new OResPAIR (a2 -> One () -> Copy (), v -> Copy ());
//         OResSET *res = new OResSET (two -> reduce (pr -> Copy ()),
// 	          		    new BResCOMP (new BResINF (v -> Copy (), IN, a2 -> Two () -> Copy ()),
// 					          SAND,
// 					          one -> reduce (pr -> Copy ())));
//        delete a2;
//        delete v;
//        delete pr;
//        return res;

//     } else {

//       char *or = one -> representation ();
//       char *tr = two -> representation ();
//       char *r = new char [strlen (or) + strlen (tr) + strlen ("Iter (, )") + 1];
//       sprintf (r, "ITER (%s, %s)", or, tr);
//       OResult *res = new OResBB (r, a -> Copy ());
//       delete [] or;
//       delete [] tr;
//       delete [] r;
//       delete a;
//       return res;
//    } 
// }

// ////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  representation
// Parameters:	()							  
//
//  "FFO: FIter (" + one -> representation () + ")"
//         								    
//////////////////////////////////////////////////////////////////////////////

char* FIter::representation ()
{
  char *or = one -> representation ();
  char *tr = two -> representation ();
  char *res = new char [strlen (or) + strlen (tr) + strlen ("ITER (, )") + 1];
  sprintf (res, "ITER (%s, %s)", or, tr);
  delete [] or;    
  delete [] tr;    
  return res;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transform
// Parameters:	(CState*)							  
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//
//////////////////////////////////////////////////////////////////////////////

NonType* FIter::transform(CState* s)
{
  Predicate* first = (Predicate*) One() -> transform(s);
  Function* second = (Function*) Two() -> transform(s);
  return new FIter(first,second,Type());
    
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  transformDeep
// Parameters:  (CState*, CIEnvStackDataType*)
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly.
//  The difference between this method and transform method is a search
//  method used in VAR classes. (Take a look at transform and transformDeep
//  methods in BVar, FVar, PVar, or OVar.)
//  Transform method searches the top frame only,
//  while TransformDeep methods searches entire stack starting from
//  a given point.
//
//////////////////////////////////////////////////////////////////////////////

NonType* FIter::transformDeep(CState* s, CIEnvStackDataType* startingPt)
{
  Predicate* first = (Predicate*) One() -> transformDeep(s, startingPt);
  Function* second = (Function*) Two() -> transformDeep(s, startingPt);
  return new FIter(first,second,Type());
    
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: compares
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for rule matching only.
// Rule matching is different from statement matching.
// Duplicated variable entry is allowed here.
//
// For example, foo: g o g --> g o g is a legal rule.
//                  -------   -------
//                    |         |_\
//                    |_\         / building part
//                      / matching part
//
// Assume that we apply this rule to a kola tree of the form "ID o ID."
// When unification occurs in a matching part, first we store first g and ID pair
// into the env stack. Then the next time we face another g (second one),
// we search the stack for the matching pair with variable name g and simply check
// if the matching pair in the stack has same structure as the matching pair we have.
// In this case, second pair is (g, ID) as well and the rule gets fired successfully.
//
// Now assume that we apply the same rule to a kola tree of the form "ID o (ID o ID)."
// In this case, the rule fails since the second pair is (g, ID o ID),
// while the first pair is (g, ID).
//
// Note that even in a case of failure, the rule itself is legal.
//
// "Duplication of a variable is legal in a rule section."
//
//////////////////////////////////////////////////////////////////////////////

int FIter::compares(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();
  
  if ((tmp_name->tag == FNAME) && (tmp_name->fn == FUN_ITER))
	{
	  // do it for first child.
	  if (one->compares(((FIter*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one))
			// child (one) is a variable.
			{
			  if ((a->IFind(((PVar*)one)->Name())) == NULL)
				{
				  // child is a new variable.
				  a->IInsert(((PVar*)one)->Name(), new CIdent((FIter*)nt,((FIter*)nt)->One(),1));
				  // Push entry into the stack.
				}
			  else
				{
				  // Child is a variable but not new one.
				  // In this case, I need to make sure that what's
				  // already in the stack and what I have now is the same.
				  if (!(((a->IFind(((PVar*)one)->Name()))->Child_Look_Up())->deepeq(((FIter*)nt)->One())))
					return_val++;
				  // match failed.
				  // else do none.
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++;
	  // compare failed
	  
	  // do it for second child.
	  if (two->compares(((FIter*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((FVar*)two)->Name())) == NULL)
				{
				  a->IInsert(((FVar*)two)->Name(), new CIdent((FIter*)nt, ((FIter*)nt)->Two(), 2));
				}
			  else 
				{
				  if (!(((a->IFind(((FVar*)two)->Name()))->Child_Look_Up())->deepeq(((FIter*)nt)->Two())))
					return_val++;
				}
			}
		}
	  else return_val++;
	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0.
  // In this case, return false(0).
  // Otherwise, return true(1).
  
  delete tmp_name;
  
  if (return_val)
	return F;
  else
	return T;
  
} 

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: comparesSTM
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for statement matching only.
// Statement matching is different from rule matching.
// No duplicated variable entry is allowed here.
//
// For example,
//             match f ! _O,
//                   f =  g o g  is a illegal statement (equation).
//                       -------
//                         |
//                         |_\
//                           / matching part
//
// Assume that we have a query tree of the form "ID o ID ! 3."
// Variable f gets bounded to "ID o ID" and we have to match "ID o ID" to "g o g."
// Matching "ID o ID" to "g o g" is perfectly legal in a rule section.
// However, it is illegal in statement section of the program.
// Remember that every variable on the right hand side of an equation must be a fresh one.
// However, after we match first g to ID, the second g is no longer fresh variable.
//
// "No duplication of a variable is legal in a statement section."
//
//////////////////////////////////////////////////////////////////////////////

int FIter::comparesSTM(NonType* nt, CState* a)
{
  int return_val = 0;
  NameUnion* tmp_name = nt->Myname();
  
  if ((tmp_name->tag == FNAME) && (tmp_name->fn == FUN_ITER))
	{
	  // do it for first child.
	  if (one->comparesSTM(((FIter*)nt)->One(), a))
		// compare succeeded.
		{
		  if (isChildVar(one))
			// child (one) is a variable.
			{
			  if ((a->IFind(((PVar*)one)->Name())) == NULL)
				{
				  // child is a new variable.
				  a->IInsert(((PVar*)one)->Name(), new CIdent((FIter*)nt,((FIter*)nt)->One(),1));
				  // Push entry into the stack.
				}
			  else 
				{
				  return_val++;
				  cout << "Error : Variable " << ((PVar*)one)->Name() << " is used twice. " << endl;
				}
			}
		  // else do none.
		  // no child is variable.
		}
	  else return_val++;
	  // compare failed
	  
	  // do it for second child.
	  if (two->comparesSTM(((FIter*)nt)->Two(), a))
		{
		  if (isChildVar(two))
			{
			  if ((a->IFind(((FVar*)two)->Name())) == NULL)
				{
				  a->IInsert(((FVar*)two)->Name(), new CIdent((FIter*)nt, ((FIter*)nt)->Two(), 2));
				}
			  else 
				{
				  return_val++;
				  cout << "Error : Variable " << ((FVar*)two)->Name() << " is used twice. " << endl;
				}
			}
		}
	  else return_val++;
	}
  else return_val++;
  // When one of it's child is a variable and faild match, return_val
  // is set to some number bigger than 0.
  // In this case, return false(0).
  // Otherwise, return true(1).
  
  delete tmp_name;
 
  if (return_val)
        return F;
  else
        return T;

} 

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  setDEPnInvDEP
// Parameters:  (char*, CState*)
//
// set dependency list (ancestor list) and inverse dependency list (descendant list)
//
//////////////////////////////////////////////////////////////////////////////

void FIter::setDEPnInvDEP(char* vn, CState* s)
{
  one->setDEPnInvDEP(vn,s);
  two->setDEPnInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  undoSetInvDEP
// Parameters:  (char*, CState*)
//
// undo changes made on descendant list (inverse dependency list).
// undo for dependency list (ancestor list) is not necessary, since
// any changes on dependency list is undone by poping up
// top frame from the environment stack.
//
//////////////////////////////////////////////////////////////////////////////

void FIter::undoSetInvDEP(char* vn, CState* s)
{
  one->undoSetInvDEP(vn,s);
  two->undoSetInvDEP(vn,s);
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  replaceChild
// Parameters:	(NonType*, int)
//
// Replace a child of given tree.  
// which_child specifies which child to replace.
// For example, when which_child is one, we replace the first child of 
// given tree.
// 
//////////////////////////////////////////////////////////////////////////////

void FIter::replaceChild(NonType* child, int which_child)
{
  if ( which_child == 1 )
	SetOne(child);
  else if ( which_child == 2 )
	SetTwo(child);
}

#ifdef _OSTORE_
//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: invoke
//  Parameters:(koObject*)
//
//  Implements the iter function former.
//
//////////////////////////////////////////////////////////////////////////////
koObject *FIter::invoke(koObject *x)
{
  void *handle = dlopen(fiterOpt, RTLD_LAZY);
  void *symbol = dlsym(handle, "callOpt");

  NonType *result = ((callOptPtr)symbol)(new OInvoke(this,x));
  dlclose(handle);
  return (koObject*)result;
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: exec
//  Parameters:(OSObject*)
//
//  Implements the iter function former.
//
//////////////////////////////////////////////////////////////////////////////

OSObject* FIter::exec(OSObject* x)
{
  OSObject *o = ((OSPair*)x)->One();
  Iterator *i = (Iterator*)((OSPair*)x)->Two();

  return new Sieve(new PCElt(One(),o->kola()),
                   new FCElt(Two(),o->kola()), i);
}

#endif // _OSTORE_
  
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  prolog
// Parameters:	()
//
// return string representation in prolog term
//
//////////////////////////////////////////////////////////////////////////////

char* FIter::prolog ()
{
  char *or = one -> prolog ();
  char *tr = two -> prolog ();
  char *res = new char [strlen (or) + strlen (tr) + strlen ("fiter(, )") + 1];
  sprintf (res, "fiter(%s, %s)", or, tr);
  delete [] or;    
  delete [] tr;    
  return res;
}

#ifdef _PROLOG_
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  sp_prolog
// Parameters:	()
//
// Since C/C++ functions only have indirect access to Prolog terms 
// via SP_term_refs, we need to convert kola tree to SP_term_refs.
// Converted kola tree will get passed to prolog runtime system.
//
// See the Sicstus prolog version 3#6 manual for more detailed info on
// the usage of SP_term_refs and other prolog supplied methods.
//
//////////////////////////////////////////////////////////////////////////////

SP_term_ref FIter::sp_prolog ()
{
  SP_term_ref or = one -> sp_prolog ();
  SP_term_ref tr = two -> sp_prolog ();
  SP_term_ref res = SP_new_term_ref();
  SP_cons_functor(res, SP_atom_from_string("fiter"),2,or,tr);
  return res;
}
#endif // _PROLOG_

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  sql
// Parameters:	(koObject *o)
//
// Return the SQL string equivalent of invoking this on o
//
/////////////////////////////////////////////////////////////////////////////

char* FIter::sql (koObject *o)
{
    NameUnion *nu;
    nu = o -> Myname ();
    if (nu -> on == OBJ_PAIR) {
	  delete nu;
	  koObject *o1 = o -> asPair () -> One ();
	  koObject *o2 = o -> asPair () -> Two ();
	  
	  OVar *x   = new OVar (0); 
	  OPair *op = new OPair (o1 -> Copy (), x);

	  char *sel = (two -> inv (op)) -> sql ();
	  char *fro = o2 -> sql ();
	  // New stuff:  Don't print out a trivial where clause
	  char *whe;
	  int flag = T;
	  NameUnion *nu = one -> Myname ();
	  if (nu -> pn == PRED_CONST) {
		delete nu;
		koBool *b = ((PConst *) one) -> One ();
		nu = b -> Myname ();
		if ((nu -> bn == BOOL_PRIM) && (((BPrimitive *) b) -> Value () == T))
		  flag = F;
	  };
	  delete nu;
	  
	  if (flag)
		whe = (one -> inv (op)) -> sql ();
	  else {
		whe = new char [1];
		sprintf (whe, "");
	  };
	  // End of New stuff

	  
	  if (strncmp (fro, "ERROR", 5) == 0) {
		delete [] sel;
		delete [] whe;
		delete op;            // Deleting op deletes x
		return fro;
	  } else if (strncmp (sel, "ERROR", 5) == 0) {
		delete [] fro;
		delete [] whe;
		delete op;           // Deleting op deletes x
		return sel;
	  } else if (strncmp (whe, "ERROR", 5) == 0) {
		delete [] fro;
		delete [] sel;
		delete op;           // Deleting op deletes x
		return whe;
	  } else if (strcmp (whe, "") == 0) {
		char *res = new char [strlen ("(select  from  )") + strlen (sel) + strlen (x -> Name ()) + strlen (fro) + 1];
		sprintf (res, "(select %s from %s %s)", sel, fro, x -> Name ());
		delete [] sel;
		delete [] fro;
		delete [] whe;
		delete op;           // Deleting op deletes x
		return res;
	  } else {
		char *res = new char [strlen ("(select  from  where )") + strlen (sel) + strlen (fro) + strlen (x -> Name ()) + strlen (whe) + 1];
		sprintf (res, "(select %s from %s %s where %s)", sel, fro, x -> Name (), whe);
		delete [] sel;
		delete [] fro;
		delete [] whe;
		delete op;           // Deleting op deletes x
		return res;
	  } 
	} else {
	  delete nu;
	  cerr << "ERROR: Calling ITER on non-Pair Object" << endl;
	  return NULL;
	}
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  oql
// Parameters:	(koObject *o)
//
// Return the OQL string equivalent of invoking this on o
//
/////////////////////////////////////////////////////////////////////////////

char* FIter::oql (koObject *o)
{
    NameUnion *nu;
    nu = o -> Myname ();
    if (nu -> on == OBJ_PAIR) {
	  delete nu;
	  koObject *o1 = o -> asPair () -> One ();
	  koObject *o2 = o -> asPair () -> Two ();
	  
	  OVar *x   = new OVar (0); 
	  OPair *op = new OPair (o1 -> Copy (), x);

	  char *sel = (two -> inv (op)) -> oql ();
	  char *fro = o2 -> oql ();
	  // New stuff:  Don't print out a trivial where clause
	  char *whe;
	  int flag = T;
	  NameUnion *nu = one -> Myname ();
	  if (nu -> pn == PRED_CONST) {
		delete nu;
		koBool *b = ((PConst *) one) -> One ();
		nu = b -> Myname ();
		if ((nu -> bn == BOOL_PRIM) && (((BPrimitive *) b) -> Value () == T))
		  flag = F;
	  };
	  delete nu;
	  
	  if (flag)
		whe = (one -> inv (op)) -> oql ();
	  else {
		whe = new char [1];
		sprintf (whe, "");
	  };
	  // End of New stuff

	  
	  if (strncmp (fro, "ERROR", 5) == 0) {
		delete [] sel;
		delete [] whe;
		delete op;            // Deleting op deletes x
		return fro;
	  } else if (strncmp (sel, "ERROR", 5) == 0) {
		delete [] fro;
		delete [] whe;
		delete op;           // Deleting op deletes x
		return sel;
	  } else if (strncmp (whe, "ERROR", 5) == 0) {
		delete [] fro;
		delete [] sel;
		delete op;           // Deleting op deletes x
		return whe;
	  } else if (strcmp (whe, "") == 0) {
		char *res = new char [strlen ("(select  from  )") + strlen (sel) + strlen (x -> Name ()) + strlen (fro) + 1];
		sprintf (res, "(select %s from %s %s)", sel, fro, x -> Name ());
		delete [] sel;
		delete [] fro;
		delete [] whe;
		delete op;           // Deleting op deletes x
		return res;
	  } else {
		char *res = new char [strlen ("(select  from  where )") + strlen (sel) + strlen (fro) + strlen (x -> Name ()) + strlen (whe) + 1];
		sprintf (res, "(select %s from %s %s where %s)", sel, fro, x -> Name (), whe);
		delete [] sel;
		delete [] fro;
		delete [] whe;
		delete op;           // Deleting op deletes x
		return res;
	  } 
	} else {
	  delete nu;
	  cerr << "ERROR: Calling ITER on non-Pair Object" << endl;
	  return NULL;
	}
}
