/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//                                                                          
//  NAME: Mitch Cherniack (mfc)					
//        Joon S. Lee (jl) 	    
//  FILE: KOFTimes.H 
//  PROJ: COKO, KOLA
//  DATE: 5/97		                                                    
//
// SYNOPSIS
//  The pairwise function former.  Given functions
//  f and g, we have:
//  
//    f x g ! [e, e'] = [f ! e, g ! e']
//      
//////////////////////////////////////////////////////////////////////////////

#ifndef KOLA_FTIMES_HEADER
#define KOLA_FTIMES_HEADER 

#include "KGlobals.H"
#include "KOFunction.H"
#include "KONameUnion.H"
#include "KONonType.H"
#include "KOOPair.H"
#include "KOOInvoke.H"

#include "PCState.H"
#include "KOFVar.H"
#include <string.h>

class FTimes : public Function
{
public:
  FTimes (Function* f1, Function* f2, TFunc * ftype = NULL);
  //  Constructor.  f1 and f2 are function arguments.

  int compares(NonType* nt, CState* a);
  // Checks for match between NonType and NonType.
  // Stores a variable and CIdent pair onto the stack.
  // This method is for rule matching only.
  // Rule matching is different from statement matching.
  // Duplicated variable entry is allowed here. 

  int comparesSTM(NonType* nt, CState* a);
  // Checks for match between two NonType trees (query tree and pattern tree).
  // Stores a variable and CIdent pair onto the stack.
  // This match method is used for matching statement.
  // No duplicated variable entry is allowed.
  
  NonType* transform(CState* s);
  //  Return coresponding Nontype object.
  //  CState is only used in VAR classes.
  //  No other classes need CState.
 
  void setDEPnInvDEP(char*, CState*);
  // set dependency list and inverse dependency list

  void undoSetInvDEP(char*, CState*);
  // undo changes made on inverse dependency list.
  // undo for dependency list is not necessary, since 
  // any changes on dependency list is undone by poping up
  // top frame from the environment stack.

  NonType* transformDeep(CState* s, CIEnvStackDataType* startingPt);
  // difference between transform and 
  // transformDeep is that the former uses
  // IFind method and later uses IFindDeep 
  // method. 

  void replaceChild(NonType* child, int which_child);
  // Replace a child of given tree.
    
  FTimes (const FTimes& other);
  //  Copy Constructor. 
  
  ~FTimes ();
  //  Destructor. 

  Function *Copy () const;
  // Virtual Copy Constructor
    
  int deepeq(NonType * other);
  // Value-based equality

  NameUnion* Myname();
  // returns myname wrapped in side of a class NameUnion.

  int numChildren() const { return 2; }
  //  Number of arguments.

  Function* One () const {return one;}
  //  What is the function that is the first argument?
  
  Function* Two () const {return two;}
  //  What is the function that is the second argument?
  
  //  OResult* reduce (OResult* a);
  //  return
  //    new OResPAIR
  //     (one -> reduce (a -> fst -> reduce (),
  //      two -> reduce (a -> snd -> reduce ())
			   
  char* prolog ();
  // return string representation in prolog term

  char* representation ();
  //  "FFO: TIMES (" + one -> representation () + 
  //  		  two -> representation () + ")"

  const char * Name() const { return namestring; }
  // Return the text representation of the name.

  void SetOne(NonType* a) { delete one; one = (Function*)a; }
  // replace existing child to new one.
 
  void SetTwo(NonType* a) { delete two; two = (Function*)a; }
  // replace existing child to new one.

  void initCurrChild() {currChild = 0;}
  // set (reset) currChild to 0.
  // This needs to be called before using nextChild() for first time.

  NonType* nextChild();
  // returns next child. (next Child = current Child + 1) 
  //If current child is the last child return NULL.

  char *oql (koObject *o);
  // Return the OQL string equivalent of invoking this on o

  koObject *inv (koObject *o);
  // Symbolic invocation of this on o.  Eventually,
  // this method will replace "invoke" presently
  // only defined with _OSTORE_ flag set

#ifdef _PROLOG_
  SP_term_ref sp_prolog ();
    // convert kola tree into prolog term and return it.
#endif // _PROLOG_ 

#ifdef _OSTORE_
  koObject* invoke(koObject* x);
  //  Implements the pairwise function former for koObjects.

  OSObject* exec(OSObject* x);
  //  Implements the pairwise function former for OSObjects.
#endif // _OSTORE_

 protected:
  Function* one;
  //  What is the function that is the first argument?
  
  Function* two;
  //  What is the function that is the 2nd argument to
  //  the former?

  int currChild;

};

# endif

