/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Joon S. Lee (jl@cs.brown.edu)
// FILE: KOPVar.C
// PROJ: COKO
// DATE: 4/96
//
//////////////////////////////////////////////////////////////////////////////

#include "KOPVar.H"

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  PVar
// Parameters:	(char* n)							  
//
//  Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

PVar::PVar (char *s, TPred * t)  : Predicate (t, PRED_VAR,s)
{
  if (s == NULL) val = NULL;
  else {
    val = new char [strlen(s) + 1];
    sprintf (val, "%s", s);
  }
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  PVar
// Parameters:	(PVar& other)							  
//
//  Copy Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

PVar::PVar (const PVar& other) : Predicate ((Predicate &) other)
{
  if (other.val == NULL) val = NULL;
  else {
    val = new char [strlen(other.val) + 1];
    sprintf (val, "%s", other.val);
  }
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~PVar
// Parameters:	()							  
//
//  Destructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

PVar::~PVar ()
{
 if (val) {delete [] val; val = NULL;}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  reduce
// Parameters:	()							  
//
/////////////////////////////////////////////////////////////////////////////

// BResult* PVar::reduce (OResult* a)
// {
//   BResBB *res = new BResBB ((char*) Name(), a -> Copy ());
//   delete a;
//   return res;
// }

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  representation
// Parameters:	()							  
//
//  "PVAR: " + name
//         								    
//////////////////////////////////////////////////////////////////////////////

char* PVar::representation ()
{

    char *res = new char [strlen (val) + strlen ("(P: )") + 1];
    sprintf (res, "(P: %s)", val);

    return res;


}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Copy
// Parameters:	()							  
//
//  return copy of this as Predicate*
//         								    
//////////////////////////////////////////////////////////////////////////////

Predicate* PVar::Copy () const
{

    return new PVar (*this);
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transform
// Parameters:	(CState*)							  
//
//  if there is a entry for this variable name in the stack
//  return the corresponding NonType else return NULL
//         								    
//////////////////////////////////////////////////////////////////////////////

NonType* PVar::transform(CState* a)
{
  if ((a->IFind(val)) != NULL)
	{
	  a->SetSuccess(T);
	  return ((Predicate*)((a->IFind(val))->Child()));
	}
  else
	{
	  a->SetSuccess(F);
	  return NULL;
	}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: partial_transform
// Parameters:	(CState*)							  
//
// similar to transform except one thing.
// If no match is found for variable,
// do not report it as an error. Instead,
// just return variable itself.
//         								    
//////////////////////////////////////////////////////////////////////////////

NonType* PVar::partial_transform(CState* a)
{
  if ((a->IFind(val)) != NULL)
	{
	  a->SetSuccess(T);
	  return (((Predicate*)((a->IFind(val))->Child())));
	}
  else
	{
	  a->SetSuccess(T);
	  return Copy();
	}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  transformDeep
// Parameters:	(CState*, CIEnvStackDataType*)							  
//
//  Return coresponding NonType object.
//  CState is only used in VAR classes.
//  No other classes need CState.
//  However, CState is kept for making transform methods uniformly. 
//  The difference between this method and transform method is a search
//  method used in VAR classes. (Take a look at transform and transformDeep
//  methods in BVar, FVar, PVar, or OVar.
//
//////////////////////////////////////////////////////////////////////////////

NonType* PVar::transformDeep(CState* a, CIEnvStackDataType* startingPt)
{
  if ((a->IFindFromHere(val, startingPt)) != NULL)
	{
	  a->SetSuccess(T);
	  return ((Predicate*)((a->IFindFromHere(val, startingPt))->Child()));
	}
  else
	{
	  a->SetSuccess(F);
	  return NULL;
	}
}

//////////////////////////////////////////////////////////////////////////////
//
//  Function Name: compares
//  Parameters:(NonType*, CState*)
//
// Checks for match between two NonType trees (query tree and pattern tree).
// Stores a variable and CIdent pair onto the stack.
// This method is for rule matching only.
// Rule matching is different from statement matching.
// Duplicated variable entry is allowed here. 
//
//////////////////////////////////////////////////////////////////////////////

int PVar::compares(NonType* nt, CState* a)
{
  // Check condition: is this right?

  NameUnion* tmp_name =  nt->Myname();
  if ((tmp_name->tag == PNAME) && (tmp_name->pn != PRED_INVALID) &&
	  (tmp_name->pn != PRED_DC) && (tmp_name->pn != PRED_VAR)) 
	{
	  delete tmp_name;
	  return T;
	}
  else
	{
	  delete tmp_name;
	  return F;
	}
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Myname
// Parameters:	()							  
//
//  Returns a union type, NameUnion. 
//  Since NonType has 5 different sub-classes, Function, Predicate, Object, 
//  Bool and Index, and those 4 different sub-classes use different 
//  enum type for their own sub-classes' names, Myname methods needs 
//  a union type of 5 different enum types.
//  Since Myname is defined virtually in KONonType, one can just call this 
//  methods with NonType* to identify the specific instance of Kola object.
//         								    
//////////////////////////////////////////////////////////////////////////////

NameUnion* PVar::Myname()
{
	NameUnion* a = new NameUnion(PNAME, PRED_VAR);
	return a;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name: directStore
// Parameters:	(CState*)							  
//
//  For variables, I only need to check whether the corresponding
//  Nontype Class is as same type as current NonType.
//
//////////////////////////////////////////////////////////////////////////////

CState* PVar::directStore(CState* a)
{
  if (a->Store() == NULL)
        {
          cout << "EOR -------- EOR ---- Store is NULL :: PVar" << endl;
          return NULL;
        }

  NameUnion* tmp_name =  (a->Store())->Myname();
  // (tmp_name->pn >=  PRED_EQUAL)) && (tmp_name->pn <= PRED_NSIMILAR))
  if ((tmp_name->tag == PNAME) && (tmp_name->pn != PRED_INVALID) &&
	  (tmp_name->pn != PRED_DC) && (tmp_name->pn != PRED_VAR)) 
	{
	  if (((a->IFind(val))) == NULL)
		{
		  a->IInsert(val, new CIdent( NULL, a->Store(), 0));
		  a->SetSuccess(T);
		}
	  else if (((a->IFind(val))) != NULL)
		{
		  if (!((a->IFind(val))->Child_Look_Up())->deepeq(a->Store()))
			a->SetSuccess(F);
		  else
			a->SetSuccess(T);
		}
	}
  else a->SetSuccess(F);
 
 delete tmp_name;
 return a;
}      

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  setInvDEP
// Parameters:  (char*, CState*)
//
// set dependency list and inverse dependency list
//
//////////////////////////////////////////////////////////////////////////////

void PVar::setDEPnInvDEP(char* vn, CState* s)
{
  if (vn == NULL)
	  s->CInsertInvDep(val);
  else
	{
	  s->IInsertInvDep(vn, val);
	  s->IInsertDep(val, vn);
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  undoSetInvDEP
// Parameters:  (char*, CState*)
//
// undo changes made on inverse dependency list.
// undo for dependency list is not necessary, since 
// any changes on dependency list is undone by poping up
// top frame from the environment stack.
//
//////////////////////////////////////////////////////////////////////////////

void PVar::undoSetInvDEP(char* vn, CState* s)
{
  s->IDeleteInvDep(val);
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  replaceChild
// Parameters:	(NonType*, int)
//
// Replace a child of given tree.  
// which_child specifies which child to replace.
// For example, when which_child is one, we replace the first child of 
// given tree.
// 
//////////////////////////////////////////////////////////////////////////////

void PVar::replaceChild(NonType* child, int which_child)
{
}

  
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  prolog
// Parameters:	()
//
// return string representation in prolog term
//
//////////////////////////////////////////////////////////////////////////////

char* PVar::prolog ()
{
  char *res = new char [strlen (val) + strlen ("V") + 1];
  sprintf (res, "V%s", val);
  return res;
}

#ifdef _PROLOG_
//////////////////////////////////////////////////////////////////////////////
//
// Function Name:  sp_prolog
// Parameters:	()
//
// Since C/C++ functions only have indirect access to Prolog terms 
// via SP_term_refs, we need to convert kola tree to SP_term_refs.
// Converted kola tree will get passed to prolog runtime system.
//
// See the Sicstus prolog version 3#6 manual for more detailed info on
// the usage of SP_term_refs and other prolog supplied methods.
//
//////////////////////////////////////////////////////////////////////////////

SP_term_ref PVar::sp_prolog ()
{
  cout << "Error @PVar::sp_prlog" << endl;
  cout << "query input should not have variable inside error!" << endl;
  return NULL;
}
#endif // _PROLOG_

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Val
// Parameters:	()
//
// Selector - return a copy of val
//         								    
//////////////////////////////////////////////////////////////////////////////

char* PVar::Val ()
{
  char *res = new char [strlen (val) + 1];
  sprintf (res, "%s", val);
  return res;
}
