/*
 * Copyright 1994, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
 */

/* -*-Mode:c++-*- */

//////////////////////////////////////////////////////////////////////////////
//
// NAME: Mitch Cherniack (mfc@cs.brown.edu)
//       Joon S. Lee (jl@cs.brown.edu)
// FILE: KOPredicate.C
// PROJ: COKO Compiler, KOLA Parcer
// DATE: 5/97
//
//////////////////////////////////////////////////////////////////////////////

#include "KOPredicate.H"
#include "KOBInvoke.H"

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Predicate
// Parameters:	(TPred *t, PredName pn, char* nm)
//
//  Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

Predicate::Predicate (TPred *t, PredName pn, char * nm)
{
  type = t;
  myname = pn;

  if (nm == NULL) namestring = NULL;
  else {
    namestring = new char [strlen(nm) + 1];
    sprintf (namestring, "%s", nm);
  }
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  Predicate
// Parameters:  (const Predicate& other)
//
//  Copy Constructor.
//         								    
//////////////////////////////////////////////////////////////////////////////

Predicate::Predicate (const Predicate& other)
{
  if (other.Type() == NULL) type = NULL;
  else type = new TPred (* (other.Type()));
  myname = other . Identity ();

  namestring = new char [strlen(other.Name()) + 1];
  sprintf (namestring, "%s", other.Name());
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  ~Predicate
// Parameters:	()
//
//  Destructor.
//         								    
//////////////////////////////////////////////////////////////////////////////
    
Predicate::~Predicate ()
{
  delete type;
  delete [] namestring;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  sql
// Parameters:	(koObject *o)
//
//  Returns SQL representation of result of invoking this on o
//         								    
//////////////////////////////////////////////////////////////////////////////

char* Predicate::sql (koObject *o)
{
  BInvoke *bi = new BInvoke (Copy (), o -> Copy ());
  char *id = bi -> representation ();

  char *res = new char [strlen ("ERROR: No SQL Equivalent of \nCalling sql on: ") + strlen (namestring) + strlen (id) + 1];
  sprintf (res, "ERROR: No SQL Equivalent of %s\nCalling sql on: %s", namestring, id);

  delete bi;
  delete [] id;
  return res;
}

//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  oql
// Parameters:	(koObject *o)
//
//  Returns OQL representation of result of invoking this on o
//         								    
//////////////////////////////////////////////////////////////////////////////

char* Predicate::oql (koObject *o)
{
  BInvoke *bi = new BInvoke (this, o);
  char *id = bi -> representation ();

  char *res = new char [strlen ("ERROR: No OQL Equivalent of \nCalling oql on: \n") + strlen (namestring) + strlen (id) + 1];
  sprintf (res, "ERROR: No OQL Equivalent of %s\nCalling oql on: %s\n", namestring, id);

  delete bi;
  delete [] id;
  return res;

}



//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  inv
// Parameters:	(koObject *o)
//
// Symbolic invocation of this on o.  Eventually,
// this method will replace "invoke" presently
// only defined with _OSTORE_ flag set
//  
//////////////////////////////////////////////////////////////////////////////

koBool* Predicate::inv (koObject* o)
{
  return new BInvoke (Copy (), o -> Copy ());
}

#ifdef _OSTORE_
//////////////////////////////////////////////////////////////////////////////
//                                                                          
// Function Name:  invoke
// Parameters:	(koObject *x)
//
//  Invokes the COKO-generated optimizer to transform
//  to a more efficient expression.	    
//////////////////////////////////////////////////////////////////////////////

koBool* Predicate::invoke(koObject* x)
{
  return new BInvoke(this, x);
}
#endif // _OSTORE_
  
